/*--------------------------------------------------------------------*/
/*    Copyright 2005 Sandia Corporation.                              */
/*    Under the terms of Contract DE-AC04-94AL85000, there is a       */
/*    non-exclusive license for use of this work by or on behalf      */
/*    of the U.S. Government.  Export of this program may require     */
/*    a license from the United States Government.                    */
/*--------------------------------------------------------------------*/

#include <sstream>
#include <string>

#include "HexBeam.hpp"
#include "hexbeam_exception.hpp"
#include "hexbeam_functions.hpp"

HexBeam::HexBeam(int W, int D,
		 Decomposition decomp,
                 int numProcs, int localProc,
                 bool print_numProcsW_numProcsD_to_cout)
  : W_(W),
    D_(D),
    decomp_(decomp),
    numProcs_(numProcs),
    numProcsW_(0),
    numProcsD_(0),
    localProc_(localProc),
    localProcW_(0),
    localProcD_(0),
    minX_(0),
    minY_(0),
    minZ_(0),
    maxX_(0),
    maxY_(0),
    maxZ_(0),
    totalNumElems_(0),
    totalNumNodes_(0),
    localNumElems_(0),
    localNumNodes_(0),
    numElemsPerSlice_(0),
    numNodesPerSlice_(0),
    numLocalSlices_(0),
    nodesPerElem_(8)
{
  totalNumElems_ = W*W*D;
  totalNumNodes_ = (W+1)*(W+1)*(D+1);
  numElemsPerSlice_ = W*W;
  numNodesPerSlice_ = (W+1)*(W+1);

  std::pair<int,int> proc_grid(1, numProcs);

  if (decomp == HexBeam::TwoD) {
    proc_grid = hexbeam::compute_num_procs_2D(numProcs, W, D);
  }

  numProcsW_ = proc_grid.first;
  numProcsD_ = proc_grid.second;

  localProcW_ = localProc%numProcsW_;
  localProcD_ = localProc/numProcsW_;

  numLocalSlices_ = D/numProcsD_;
  int remainder = D%numProcsD_;

  if (localProcD_ < remainder) {
    ++numLocalSlices_;
  }

  int Wrows = W/numProcsW_;

  localNumNodes_ = (Wrows+1)*(W+1)*(numLocalSlices_+1);

  localNumElems_ = (numElemsPerSlice_/numProcsW_)*numLocalSlices_;

  minX_ = 0;
  maxX_ = W;
  minY_ = localProcW_*Wrows;
  maxY_ = minY_ + Wrows;
  minZ_ = localProcD_*numLocalSlices_;
  if (localProcD_ >= remainder) {
    minZ_ += remainder;
  }

  maxZ_ = minZ_ + numLocalSlices_;

  if (print_numProcsW_numProcsD_to_cout && localProc == 0) {
    std::cout << "Decomposition: " << to_string(decomp_)
       << ", numProcsTotal: " << numProcs_ << ", numProcsW: "
        << numProcsW_ << ", numProcsD: " << numProcsD_ << std::endl;
  }
}

HexBeam::~HexBeam()
{
}

void checkRange(int item, int min_valid, int max_valid,
                const std::string& msgprefix)
{
  if (item < min_valid || item > max_valid) {
    std::ostringstream osstr;
    osstr << msgprefix <<item<<" out of range"
       << "; must be in the range ["<<min_valid<<" .. "<<max_valid<<"].";
    std::string msg = osstr.str();
    throw hexbeam::exception(msg);
  }
}

int HexBeam::getLocalElemIDs(std::vector<int>& elemIDs)
{
  elemIDs.resize(localNumElems_);

  unsigned offset = 0;
  for(int z=minZ_; z<maxZ_; ++z) {
    for(int y=minY_; y<maxY_; ++y) {
      for(int x=minX_; x<maxX_; ++x) {
        elemIDs[offset++] = 1 + W_*W_*z + W_*y + x;
      }
    }
  }

  return(0);
}

int HexBeam::getLocalNodeIDs(std::vector<int>& nodeIDs)
{
  nodeIDs.resize(localNumNodes_);

  unsigned offset = 0;
  for(int z=minZ_; z<=maxZ_; ++z) {
    for(int y=minY_; y<=maxY_; ++y) {
      for(int x=minX_; x<=maxX_; ++x) {
        nodeIDs[offset++] = getNodeID(x,y,z);
      }
    }
  }

  return(0);
}

int HexBeam::getElemConnectivity(int elemID, int* nodeIDs)
{
  checkRange(elemID, 1, totalNumElems_,
             "HexBeam::getElemConnectivity: elemID ");

  int X=0, Y=0, Z=0;
  getElemCoords(elemID, X, Y, Z);
  int firstElemNode = getNodeID(X, Y, Z);

  nodeIDs[0] = firstElemNode;
  nodeIDs[1] = firstElemNode+1;
  nodeIDs[2] = firstElemNode+W_+1;
  nodeIDs[3] = nodeIDs[2]+1;

  nodeIDs[4] = nodeIDs[0]+numNodesPerSlice_;
  nodeIDs[5] = nodeIDs[1]+numNodesPerSlice_;
  nodeIDs[6] = nodeIDs[2]+numNodesPerSlice_;
  nodeIDs[7] = nodeIDs[3]+numNodesPerSlice_;

  return(0);
}

int HexBeam::getNumBCNodes()
{
  int numBCNodes = (numLocalSlices_+1)*(W_+1);
  return( numBCNodes );
}

int HexBeam::getBCNodes(int numNodes, int* nodeIDs)
{
  if (numNodes != getNumBCNodes()) {
    return(-1);
  }

  int firstBCNode = W_;

  for(int i=0; i<numNodes; ++i) {
    nodeIDs[i] = firstBCNode + W_+1;
  }

  return(0);
}

int HexBeam::getNumSharedNodes()
{
  //if running on only 1 proc, or if maxZ_ not greater than minZ_,
  //then there are no shared nodes.
  if (numProcs_ <= 1 || maxZ_ <= minZ_) return(0);

  int numSharedNodes = 0;

  //since HexBeam can only be distributed by cutting across the Z dimension
  //or across the Z dimension and one of the W dimensions, we know that
  //minX_ == 0 and maxX_ == W_ always.
  //So calculation of shared nodes only requires checking whether we
  //have neighbors in the Y and Z dimensions.

  if (minY_ > 0) {
    numSharedNodes += (maxX_-minX_)*(maxZ_-minZ_);
  }

  if (maxY_ < W_) {
    numSharedNodes += (maxX_-minX_)*(maxZ_-minZ_);
  }

  if (minZ_ > 0) {
    numSharedNodes += (maxX_-minX_)*(maxY_-minY_);
  }

  if (maxZ_ < D_) {
    numSharedNodes += (maxX_-minX_)*(maxY_-minY_);
  }

  return(numSharedNodes);
}

void insert_node_proc(int node, int proc,
                      std::map<int,std::set<int>*>& node_proc_map)
{
  typedef std::map<int,std::set<int>*> mapset;

  std::set<int>* procset = NULL;
  mapset::iterator iter = node_proc_map.find(node);
  if (iter == node_proc_map.end()) {
    procset = new std::set<int>;
    node_proc_map.insert(std::make_pair(node,procset));
  }
  else {
    procset = iter->second;
  }

  procset->insert(proc);
}

int HexBeam::getSharedNodes(std::vector<int>& sharedNodes,
                            std::vector<int>& numSharingProcsPerNode,
                            std::vector<int>& sharingProcs)
{
  sharedNodes.clear();
  numSharingProcsPerNode.clear();
  sharingProcs.clear();

  //if running on only 1 proc, or if maxZ_ not greater than minZ_,
  //then there are no shared nodes.
  if (numProcs_ <= 1 || maxZ_ <= minZ_) return(0);

  typedef std::map<int,std::set<int>*> mapset;
  mapset sharedNodeProcMap;

  if (minY_ > 0) {
    for(int z=minZ_; z<=maxZ_; ++z) {
      for(int x=minX_; x<=maxX_; ++x) {
        int node = getNodeID(x, minY_, z);

        insert_node_proc(node, localProc_-1, sharedNodeProcMap);

        if (z==minZ_ && minZ_ > 0) {
          insert_node_proc(node, localProc_-numProcsW_-1, sharedNodeProcMap);
        }

        if (z==maxZ_ && maxZ_ < D_) {
          insert_node_proc(node, localProc_+numProcsW_-1, sharedNodeProcMap);
        }
      }
    }
  }

  if (maxY_ < W_) {
    for(int z=minZ_; z<=maxZ_; ++z) {
      for(int x=minX_; x<=maxX_; ++x) {
        int node = getNodeID(x, maxY_, z);

        insert_node_proc(node, localProc_+1, sharedNodeProcMap);

        if (z==minZ_ && minZ_ > 0) {
          insert_node_proc(node, localProc_-numProcsW_+1, sharedNodeProcMap);
        }

        if (z==maxZ_ && maxZ_ < D_) {
          insert_node_proc(node, localProc_+numProcsW_+1, sharedNodeProcMap);
        }
      }
    }
  }

  if (minZ_ > 0) {
    for(int y=minY_; y<=maxY_; ++y) {
      for(int x=minX_; x<=maxX_; ++x) {
        int node = getNodeID(x, y, minZ_);

        insert_node_proc(node, localProc_-numProcsW_, sharedNodeProcMap);

        if (y==minY_ && minY_ > 0) {
          insert_node_proc(node, localProc_-numProcsW_-1, sharedNodeProcMap);
        }

        if (y==maxY_ && maxY_ < W_) {
          insert_node_proc(node, localProc_-numProcsW_+1, sharedNodeProcMap);
        }
      }
    }
  }

  if (maxZ_ < D_) {
    for(int y=minY_; y<=maxY_; ++y) {
      for(int x=minX_; x<=maxX_; ++x) {
        int node = getNodeID(x, y, maxZ_);

        insert_node_proc(node, localProc_+numProcsW_, sharedNodeProcMap);

        if (y==minY_ && minY_ > 0) {
          insert_node_proc(node, localProc_+numProcsW_-1, sharedNodeProcMap);
        }

        if (y==maxY_ && maxY_ < W_) {
          insert_node_proc(node, localProc_+numProcsW_+1, sharedNodeProcMap);
        }
      }
    }
  }

  int numSharedNodes = sharedNodeProcMap.size();
  sharedNodes.resize(numSharedNodes);
  numSharingProcsPerNode.resize(numSharedNodes);

  mapset::iterator
    iter = sharedNodeProcMap.begin(), end = sharedNodeProcMap.end();

  unsigned totalsize = 0;
  for(; iter!=end; ++iter) {
    totalsize += iter->second->size();
  }

  sharingProcs.resize(totalsize);

  iter = sharedNodeProcMap.begin();
  unsigned i=0;
  unsigned offset = 0;
  for(; iter!=end; ++iter) {
    int node = iter->first;
    std::set<int>* procset = iter->second;
    sharedNodes[i] = node;
    numSharingProcsPerNode[i++] = procset->size();

    std::set<int>::iterator
      s_iter = procset->begin(), s_end = procset->end();
    for(; s_iter!=s_end; ++s_iter) {
      sharingProcs[offset++] = *s_iter;
    }

    delete procset;
  }

  return(0);
}

