#ifndef _HexBeam_hpp_
#define _HexBeam_hpp_

/*--------------------------------------------------------------------*/
/*    Copyright 2005 Sandia Corporation.                              */
/*    Under the terms of Contract DE-AC04-94AL85000, there is a       */
/*    non-exclusive license for use of this work by or on behalf      */
/*    of the U.S. Government.  Export of this program may require     */
/*    a license from the United States Government.                    */
/*--------------------------------------------------------------------*/

#include <vector>

/**
HexBeam generates data that represents a beam of hex-8 elements.
More precisely, it's a cube that can be elongated in one dimension, forming
an arbitrarily long beam.

The number of elements in the beam is WxWxD. (W for 'width', D for 'depth')
The number of nodes in the beam is (W+1)x(W+1)x(D+1).

In 3D space the beam's "long" dimension (D) extends along the Z axis.
Each element is a unit cube.
Nodes are numbered starting with node 1 at the origin.
The x,y,z coordinates of a few specific nodes are:
node 1:                 0,0,0
node W+1:               W,0,0
node (W+1)*(W+1):       W,W,0
node (W+1)*(W+1)*(D+1): W,W,D

Node numbers increase with stride 1 along the X axis, stride W+1 along
the Y axis, and stride (W+1)*(W+1) along the Z axis.

Element numbers increase in a similar order: stride 1 along the X axis,
stride W along the Y axis, and stride W*W along the Z axis.

The 8 nodes connected to each element are always listed in the same order.
Taking element 1 as an example, the connected nodes are:

local-node   global-node            coords
   0           1                    0,0,0
   1           2                    1,0,0
   2           2+W+1                1,1,0
   3           1+W+1                0,1,0
   4           (W+1)*(W+1)+1        0,0,1
   5           (W+1)*(W+1)+2        1,0,1
   6           (W+1)*(W+1)+2+W+1    1,1,1
   7           (W+1)*(W+1)+1+W+1    0,1,1

Given a global element number elem, the coordinates of that element's
"local-node 0" are found using integer arithmetic:
X: (elem-1)%W
Y: ((elem-1)%(W*W))/W
Z: (elem-1)/(W*W)

Given a global node number n, the coordinates of that node are:
X: (n-1)%(W+1)
Y: ((n-1)%((W+1)*(W+1)))/(W+1)
Z: (n-1)/((W+1)*(W+1))

Given coordinates x,y,z, the node-number at that position is:
  n = 1 + (W+1)*(W+1)*z + (W+1)*y + x

Given coordinates x,y,z, the element-number for which local-node 0 lies at
that position is:
  elem = 1 + W*W*z + W*y + x

In parallel runs, the beam is divided such that each processor has a
portion of the elements.

If the "OneD" decomposition option is used, then each processor has roughly
D/numprocs "slices" of the beam (sliced across the D dimension), and each
slice has WxW elements.

If the "TwoD" decomposition option is used, then the beam is sliced across
one of the W dimensions as well as across the D dimension.

Elements that are separated by processor boundaries share common nodes, which
are "shared nodes" in the global mesh.

 */
class HexBeam {
 public:
  enum Decomposition { OneD = 1, TwoD = 2 };

  HexBeam(int W, int D,
          Decomposition decomp,
          int numProcs, int localProc,
          bool print_numProcsW_numProcsD_to_cout = false);

  virtual ~HexBeam();

  virtual int numNodesPerElem() const { return( nodesPerElem_ ); }

  virtual int numLocalElems() const { return( localNumElems_ ); }

  virtual int totalNumElems() const { return( totalNumElems_ ); }

  virtual int numLocalNodes() const { return( localNumNodes_ ); }

  virtual int totalNumNodes() const { return( totalNumNodes_ ); }

  virtual int minLocalX() const { return( minX_ ); }
  virtual int minLocalY() const { return( minY_ ); }
  virtual int minLocalZ() const { return( minZ_ ); }

  virtual int maxLocalX() const { return( maxX_ ); }
  virtual int maxLocalY() const { return( maxY_ ); }
  virtual int maxLocalZ() const { return( maxZ_ ); }

  /** Fill vector with IDs of all local elements. There is no ghosting,
    so all local elements are locally owned.
  */
  virtual int getLocalElemIDs(std::vector<int>& elemIDs);

  /** Fill vector with IDs of all local nodes. These include locally
    owned nodes as well as shared-but-not-owned nodes.
  */
  virtual int getLocalNodeIDs(std::vector<int>& nodeIDs);

  virtual int getElemConnectivity(int elemID, int* nodeIDs);

  virtual int getNumBCNodes();

  virtual int getBCNodes(int numNodes, int* nodeIDs);

  virtual int getNumSharedNodes();

  virtual int getSharedNodes(std::vector<int>& sharedNodes,
                             std::vector<int>& numSharingProcsPerNode,
                             std::vector<int>& sharingProcs);

  virtual int getNumCRs() { return( 0 ); }

  virtual int getNumNodesPerCR() { return( 0 ); }

  virtual int getCRNodes(int** nodeIDs) { (void)nodeIDs; return(0); }

  void getElemCoords(int elemID, int& X, int& Y, int& Z);
  void getNodeCoords(int nodeID, int& X, int& Y, int& Z);
  void getNodeCoords(int nodeID, double& X, double& Y, double& Z);
  void getNodeCoords(int nodeID, double* coords);
  void getNodeCoords(const std::vector<int>& nodeIDs,
                     std::vector<double>& coords);

  int getNodeID(int X, int Y, int Z);

 protected:
  int W_;
  int D_;
  Decomposition decomp_;
  int numProcs_;
  int numProcsW_;
  int numProcsD_;
  int localProc_;
  int localProcW_;
  int localProcD_;

  int minX_;
  int minY_;
  int minZ_;

  int maxX_;
  int maxY_;
  int maxZ_;

  int totalNumElems_;
  int totalNumNodes_;
  int localNumElems_;
  int localNumNodes_;

  int numElemsPerSlice_;
  int numNodesPerSlice_;
  int numLocalSlices_;

  int nodesPerElem_;
};

inline
void HexBeam::getElemCoords(int elemID, int& X, int& Y, int& Z)
{
  X = (elemID-1)%W_;
  Y = ((elemID-1)%(W_*W_))/W_;
  Z = (elemID-1)/(W_*W_);
}

inline
int HexBeam::getNodeID(int X, int Y, int Z)
{
  return 1 + numNodesPerSlice_*Z + (W_+1)*Y + X;
}

inline
void HexBeam::getNodeCoords(int nodeID, int& X, int& Y, int& Z)
{
  const int Wp1 = W_+1;
  const int n = nodeID-1;
  X = n%Wp1;
  Y = (n%(Wp1*Wp1))/Wp1;
  Z = n/(Wp1*Wp1);
}

inline
void HexBeam::getNodeCoords(int nodeID, double& X, double& Y, double& Z)
{
  const int Wp1 = W_+1;
  const int n = nodeID-1;
  X = n%Wp1;
  Y = (n%(Wp1*Wp1))/Wp1;
  Z = n/(Wp1*Wp1);
}

inline
void HexBeam::getNodeCoords(int nodeID, double* coords)
{
  const int Wp1 = W_+1;
  const int n = nodeID-1;
  coords[0] = n%Wp1;
  coords[1] = (n%(Wp1*Wp1))/Wp1;
  coords[2] = n/(Wp1*Wp1);
}

inline
void HexBeam::getNodeCoords(const std::vector<int>& nodeIDs,
                   std::vector<double>& coords)
{
  coords.resize(nodeIDs.size()*3);
  double* ptr = &coords[0];
  for(unsigned i=0; i<nodeIDs.size(); ++i) {
    getNodeCoords(nodeIDs[i], ptr);
    ptr += 3;
  }
}

inline
const char* to_string(HexBeam::Decomposition decomp)
{
  switch(decomp) {
  case HexBeam::OneD:
    return "HexBeam::OneD"; break;
  case HexBeam::TwoD:
    return "HexBeam::TwoD"; break;
  default:
  return "unrecognized HexBeam::Decomposition value";
  }

  return "unrecognized HexBeam::Decomposition value";
}

#endif // _HexBeam_hpp_

