/*--------------------------------------------------------------------*/
/*    Copyright 2005 Sandia Corporation.                              */
/*    Under the terms of Contract DE-AC04-94AL85000, there is a       */
/*    non-exclusive license for use of this work by or on behalf      */
/*    of the U.S. Government.  Export of this program may require     */
/*    a license from the United States Government.                    */
/*--------------------------------------------------------------------*/

#include <string>
#include <sstream>

#include <hexbeam_functions.hpp>
#include <HexBeamCR.hpp>

HexBeamCR::HexBeamCR(int W, int D,
		 Decomposition decomp, int numProcs, int localProc)
  : HexBeam(W, D, decomp, numProcs, localProc),
    firstLocalElem_(0),
    firstLocalNode_(0)
{
  totalNumNodes_ = (W+1)*(W+1)*(D+1) + (W+1)*(W+1)*(2*numProcs_-1);

  if (decomp != HexBeam::OneD) {
    std::ostringstream osstr;
    osstr << "HexBeamCR: invalid decomp option: " << to_string(decomp)
       <<", HexBeamCR only allows HexBeam::TwoD.";
    std::string str = osstr.str();
    throw hexbeam::exception(str);
  }

  numLocalSlices_ = D/numProcs;
  int remainder = D%numProcs;
  firstLocalSlice_ = localProc_*numLocalSlices_;

  if (D < numProcs) {
    std::ostringstream osstr;
    osstr << "HexBeamCR ERROR: D < numProcs. D ("<<D<<") must be greater"
         << " than or equal to numProcs ("<<numProcs<<").";
    std::string str = osstr.str();
    throw hexbeam::exception(str);
  }

  if (localProc < remainder) {
    ++numLocalSlices_;
    ++firstLocalSlice_;
  }

  localNumNodes_ = numNodesPerSlice_*(numLocalSlices_+2);

  localNumElems_ = numElemsPerSlice_*numLocalSlices_;

  if (localProc > 0) {
    firstLocalElem_ = localProc*numLocalSlices_*numElemsPerSlice_;

    firstLocalNode_ = localProc*localNumNodes_;

    if (remainder <= localProc && remainder > 0) {
      firstLocalElem_ += remainder*numElemsPerSlice_;
      firstLocalNode_ += remainder*numNodesPerSlice_;
    }
  }

  localCRslice_ = firstLocalSlice_ + numLocalSlices_/2;
  numLocalCRs_ = numNodesPerSlice_;
  if (localProc_ < numProcs_-1) {
    numLocalCRs_ += numNodesPerSlice_;
  }

  numNodesPerCR_ = 2;
}

HexBeamCR::~HexBeamCR()
{
}

int HexBeamCR::getElemConnectivity(int elemID, int* nodeIDs)
{
  if (elemID < firstLocalElem_ || elemID > firstLocalElem_+localNumElems_) {
    return(-1);
  }

  int whichGlobalSlice = elemID/numElemsPerSlice_;
  int elemX = elemID%W_;
  int elemY = (elemID%(W_*W_))/W_;
  int firstElemNode = (whichGlobalSlice + localProc_*2)*numNodesPerSlice_
                     + elemY*(W_+1) + elemX;

  if (whichGlobalSlice >= localCRslice_) {
    firstElemNode += numNodesPerSlice_;
  }

  nodeIDs[0] = firstElemNode;
  nodeIDs[1] = firstElemNode+1;
  nodeIDs[2] = firstElemNode+W_+1;
  nodeIDs[3] = nodeIDs[2]+1;

  nodeIDs[4] = nodeIDs[0]+numNodesPerSlice_;
  nodeIDs[5] = nodeIDs[1]+numNodesPerSlice_;
  nodeIDs[6] = nodeIDs[2]+numNodesPerSlice_;
  nodeIDs[7] = nodeIDs[3]+numNodesPerSlice_;

  return(0);
}

int HexBeamCR::getCRNodes(int** nodeIDs)
{
  int offset = 0;
  int firstCRnode = (localCRslice_+localProc_*2)*numNodesPerSlice_;
  int i;
  for(i=0; i<numNodesPerSlice_; ++i) {
    nodeIDs[offset][0] = firstCRnode+i;
    nodeIDs[offset++][1] = firstCRnode+i+numNodesPerSlice_;
  }

  if (localProc_ >= numProcs_-1) return(0);

  int nextCRnode = firstLocalNode_ + localNumNodes_ - numNodesPerSlice_;
  for(i=0; i<numNodesPerSlice_; ++i) {
    nodeIDs[offset][0] = nextCRnode+i;
    nodeIDs[offset++][1] = nextCRnode+i+numNodesPerSlice_;    
  }

  return(0);
}

int HexBeamCR::getNumBCNodes()
{
  int numBCNodes = (numLocalSlices_+1)*(W_+1);
  return( numBCNodes );
}

int HexBeamCR::getBCNodes(int numNodes, int* nodeIDs)
{
  if (numNodes != getNumBCNodes()) {
    return(-1);
  }

  int firstBCNode = firstLocalNode_ + W_;

  for(int i=0; i<numNodes; ++i) {
    nodeIDs[i] = firstBCNode + W_+1;
  }

  return(0);
}

int HexBeamCR::getNumSharedNodes()
{
  if (numProcs_ < 2) return(0);

  int numSharedNodes = numNodesPerSlice_;
  if (localProc_ > 0 && localProc_ < numProcs_-1) {
    numSharedNodes += numNodesPerSlice_;
  }

  return(numSharedNodes);
}

int HexBeamCR::getSharedNodes(std::vector<int>& sharedNodes,
			      std::vector<int>& numSharingProcsPerNode,
			      std::vector<int>& sharingProcs)
{
  if (numProcs_ < 2) return(0);

  int numShared = getNumSharedNodes();
  sharedNodes.resize(numShared);
  numSharingProcsPerNode.resize(numShared);
  sharingProcs.resize(numShared);

  int firstSharedNode = firstLocalNode_;
  unsigned offset = 0;
  if (localProc_ > 0) {
    for(int i=0; i<numNodesPerSlice_; ++i) {
      sharedNodes[offset] = firstSharedNode+i;
      numSharingProcsPerNode[offset] = 1;
      sharingProcs[offset++] = localProc_-1;
    }
  }

  firstSharedNode = firstLocalNode_+numNodesPerSlice_*(numLocalSlices_+2);
  if (localProc_ < numProcs_ - 1) {
    for(int i=0; i<numNodesPerSlice_; ++i) {
      sharedNodes[offset] = firstSharedNode+i;
      numSharingProcsPerNode[offset] = 1;
      sharingProcs[offset++] = localProc_+1;
    }
  }

  return(0);
}

