/*--------------------------------------------------------------------*/
/*    Copyright 2005 Sandia Corporation.                              */
/*    Under the terms of Contract DE-AC04-94AL85000, there is a       */
/*    non-exclusive license for use of this work by or on behalf      */
/*    of the U.S. Government.  Export of this program may require     */
/*    a license from the United States Government.                    */
/*--------------------------------------------------------------------*/

//
// This is a simple program to exercise FEI classes for the
// purposes of testing, code tuning and scaling studies.
//

#include "fei_iostream.hpp" 

//Including the header fei_base.hpp pulls in the declaration for
//various classes in the 'fei' namespace.

#include "fei_base.hpp"
#include <fei_mpi.h>

//Make provision for using any one of several solver libraries. This is
//handled by the code in LibraryFactory.{hpp,cpp}.

#include "LibraryFactory.hpp"


//include some headers which are used for
//setting up the data for the hex-beam example problem.

#include "HexBeam.hpp"
#include "HexBeamCR.hpp"
#include "hexbeam_functions.hpp"

#include <fei_ParameterSet.hpp>
#include <Teuchos_ParameterList.hpp>

//fei_test_utils.hpp declares things like the intialize_mpi function.

#include "fei_test_utils.hpp"

void copy_teuchos_parameterlist(const Teuchos::ParameterList& paramlist,
                                fei::ParameterSet& paramset);
//
//Include definitions of macros like 'CHK_ERR' to call functions and check
//the return code.
//
#undef fei_file
#define fei_file "beam.cpp"
#include "fei_ErrMacros.hpp"

//==============================================================================
//Here's the main...
//==============================================================================
int beam(MPI_Comm comm, const Teuchos::ParameterList& parameterlist)
{
  int numProcs = 1;
  MPI_Comm_size(comm, &numProcs);
  int localProc = 0;
  MPI_Comm_rank(comm, &localProc);

  double start_time = fei::utils::cpu_time();

  fei::ParameterSet params;
  copy_teuchos_parameterlist(parameterlist, params);


  std::string solverName;
  std::string datasource;
  std::string constraintform;
  int W = 0;
  int D = 0;
  int DofPerNode = 0;
  std::string decomp("HexBeam::OneD");

  try {
    const std::string csolverName = parameterlist.get<std::string>("SOLVER_LIBRARY");
    const std::string cdatasource = parameterlist.get<std::string>("DATA_SOURCE");
    const int cW = parameterlist.get<int>("W");
    const int cD = parameterlist.get<int>("D");
    
    const int cDofPerNode = parameterlist.get<int>("DofPerNode");

    solverName = csolverName;
    datasource = cdatasource;
    W = cW;
    D = cD;
    DofPerNode = cDofPerNode;
  }
  catch (...) {
    //macros FEI_CERR, FEI_COUT and FEI_ENDL are aliases for
    //std::cout and std::endl, defined in fei_iostream.hpp.
    FEI_CERR << "Failed to find one or more required parameters."
	     << FEI_ENDL << "Required parameters:"<<FEI_ENDL
	     << "SOLVER_LIBRARY" << FEI_ENDL
	     << "DATA_SOURCE" << FEI_ENDL
	     << "W" << FEI_ENDL
             << "D" << FEI_ENDL
             << "DofPerNode" << FEI_ENDL;

    return(-1);
  }

  try {
    const std::string cconstraintform = parameterlist.get<std::string>("CONSTRAINT_FORM");
    const std::string cdecomp = parameterlist.get<std::string>("DECOMP");
    constraintform = cconstraintform;
    decomp = cdecomp;
  }
  catch (...) {
    //do nothing, these are not required parameters.
  }

  bool slave_constraints = (constraintform == "slaves");

  HexBeam::Decomposition decmp = HexBeam::OneD;
  if (decomp == "TwoD" || decomp == "HexBeam::TwoD" ||
      decomp == "TWOD" || decomp == "HexBeam::TWOD") {
    decmp = HexBeam::TwoD;
  }

  HexBeam* hex_beam_ptr = NULL;
  if (datasource == "HexBeam") {
    try {
      hex_beam_ptr = new HexBeam(W, D, decmp, numProcs, localProc, true);
    }
    catch(hexbeam::exception& exc) {
      if (localProc == 0) {
        FEI_CERR << "HexBeam constructor ERROR: "
          << exc.what() << FEI_ENDL;
      }
      return(-1);
    }
  }
  else {
    if (decmp == HexBeam::TwoD) {
      FEI_COUT << "Decomposition option HexBeam::TwoD not valid if "
        << "CONSTRAINT_FORM==slaves. Using HexBeam::OneD."<<FEI_ENDL;
    }

    hex_beam_ptr = new HexBeamCR(W, D, HexBeam::OneD, numProcs, localProc);
  }

  HexBeam& hex_beam = *hex_beam_ptr;

  if (localProc == 0) {
    int numCRs = (W+1)*(W+1)* ((numProcs*2)-1);
    if (hex_beam.getNumCRs() < 1) numCRs = 0;
    FEI_COUT << FEI_ENDL;
    FEI_COUT << "========================================================\n";
    FEI_COUT << "FEI version: " << fei::utils::version() << FEI_ENDL;
    FEI_COUT << "--------------------------------------------------------\n";
    FEI_COUT << "num-procs: " << numProcs << "\n";
    FEI_COUT << "Size W: " << W << " (num-elements-along-side-of-cube)\n";
    FEI_COUT << "Size D: " << D << " (num-elements-along-depth-of-cube)\n";
    FEI_COUT << "DOF per node: " << DofPerNode <<"\n";
    FEI_COUT << "Decomposition: " << to_string(decmp) << "\n";
    FEI_COUT << "Num local  elements: " << hex_beam.numLocalElems() << "\n";
    FEI_COUT << "Num global elements: " << hex_beam.totalNumElems() << "\n";
    FEI_COUT << "Num local  DOF: " << hex_beam.numLocalNodes()*DofPerNode << "\n";
    FEI_COUT << "Num global DOF: " << hex_beam.totalNumNodes()*DofPerNode << "\n";
    FEI_COUT << "Num global CRs: " << numCRs << "\n";
    FEI_COUT << "========================================================" 
	 << FEI_ENDL;
  }

  double start_init_time = fei::utils::cpu_time();

  bool create_factory_failed = false;
  fei::SharedPtr<fei::Factory> factory;
  try {
    factory = fei::create_fei_Factory(comm, solverName.c_str());
  }
  catch (std::runtime_error& exc) {
    create_factory_failed = true;
  }

  if (factory.get() == NULL || create_factory_failed) {
    FEI_COUT << "fei::Factory allocation failed or solver-library ("
            <<solverName<<" not available." << FEI_ENDL;

    return(-1);
  }

  factory->parameters(params);

  fei::SharedPtr<fei::VectorSpace> nodeSpace =
    factory->createVectorSpace(comm, "beam");

  fei::SharedPtr<fei::VectorSpace> dummy;
  fei::SharedPtr<fei::MatrixGraph> matrixGraph =
    factory->createMatrixGraph(nodeSpace, dummy, "beam");

  //load some solver-control parameters.
  nodeSpace->setParameters(params);
  matrixGraph->setParameters(params);

  int fieldID = 0;
  int fieldSize = DofPerNode;
  int nodeIDType = 0;
  int crIDType = 1;

  nodeSpace->defineFields( 1, &fieldID, &fieldSize );
  nodeSpace->defineIDTypes(1, &nodeIDType );
  nodeSpace->defineIDTypes(1, &crIDType );

  //hexbeam::init_elem_connectivities demonstrates the process
  //of initializing element-to-node connectivities on the fei::MatrixGraph.
  CHK_ERR( hexbeam::init_elem_connectivities( matrixGraph.get(), hex_beam ) );

  //hexbeam::init_shared_nodes demonstrates the process of
  //initializing shared nodes on the fei::MatrixGraph.
  CHK_ERR( hexbeam::init_shared_nodes( matrixGraph.get(), hex_beam ) );

  int firstLocalCRID = 0;
  if (slave_constraints) {
    CHK_ERR( hexbeam::
	     init_slave_constraints( matrixGraph.get(), hex_beam, DofPerNode) );
  }
  else {
    CHK_ERR( hexbeam::
	     init_constraints( matrixGraph.get(), hex_beam, localProc, firstLocalCRID) );
  }

  CHK_ERR( matrixGraph->initComplete() );

  double fei_init_time = fei::utils::cpu_time() - start_init_time;

  if (localProc == 0) {
    FEI_COUT.setf(IOS_FIXED, IOS_FLOATFIELD);
    FEI_COUT << "Initialization cpu time:   " << fei_init_time << FEI_ENDL;
  }

  //Now the initialization phase is complete. Next we'll do the load phase,
  //which for this problem just consists of loading the element data
  //(element-wise stiffness arrays and load vectors) and the boundary
  //condition data.

  double fei_creatematrix_start_time = fei::utils::cpu_time();

  fei::SharedPtr<fei::Matrix> mat = factory->createMatrix(matrixGraph);

  double fei_creatematrix_time = fei::utils::cpu_time() - fei_creatematrix_start_time;
  if (localProc == 0) {
    FEI_COUT.setf(IOS_FIXED, IOS_FLOATFIELD);
    FEI_COUT << "Create-Matrix cpu time:   " << fei_creatematrix_time << FEI_ENDL;
  }

  double start_load_time = fei::utils::cpu_time();

  fei::SharedPtr<fei::Vector> solnVec = factory->createVector(matrixGraph, true);
  fei::SharedPtr<fei::Vector> rhsVec  = factory->createVector(matrixGraph);
  fei::SharedPtr<fei::LinearSystem> linSys =
    factory->createLinearSystem(matrixGraph);

  CHK_ERR( linSys->parameters(params));

  linSys->setMatrix(mat);
  linSys->setSolutionVector(solnVec);
  linSys->setRHS(rhsVec);

  CHK_ERR( hexbeam::
	   load_elem_data(matrixGraph.get(), mat.get(), rhsVec.get(), hex_beam, DofPerNode) );

  //temporarily disable boundary-conditions
  //  CHK_ERR( load_BC_data(linSys, hex_beam, DofPerNode) );

  if (!slave_constraints) {
    CHK_ERR( hexbeam::
	     load_constraints(linSys.get(), hex_beam, DofPerNode, firstLocalCRID) );
  }

  CHK_ERR( linSys->loadComplete() );

  double fei_load_time = fei::utils::cpu_time() - start_load_time;

  if (localProc == 0) {
    //IOS macros are defined in fei_stdinc.h
    FEI_COUT.setf(IOS_FIXED, IOS_FLOATFIELD);
    FEI_COUT << "Coef. loading cpu time:    " << fei_load_time << FEI_ENDL;
    FEI_COUT << "Total assembly wall time:   "
	 << fei_init_time + fei_creatematrix_time + fei_load_time << FEI_ENDL;
  }

  //
  //now the load phase is complete, so we're ready to launch the underlying
  //solver and solve Ax=b
  //

  //First, check whether the params (set from the input .i file) specify
  //a "Trilinos_Solver" string (which may be Amesos...). If not, it won't
  //matter but if so, the factory may switch based on this value, when
  //creating the solver object instance.

  std::string solver_name_value;
  params.getStringParamValue("Trilinos_Solver", solver_name_value);

  const char* charptr_solvername =
    solver_name_value.empty() ? NULL : solver_name_value.c_str();

  fei::SharedPtr<fei::Solver> solver = factory->createSolver(charptr_solvername);

  int status;
  int itersTaken = 0;

  if (localProc==0) FEI_COUT << "solve..." << FEI_ENDL;
  double start_solve_time = fei::utils::cpu_time();

  int err = solver->solve(linSys.get(),
			  NULL, //preconditioningMatrix
			  params,
			  itersTaken,
			  status);

  double solve_time = fei::utils::cpu_time()-start_solve_time;

  if (err!=0 && localProc==0) {
    FEI_COUT << "solve returned err: " << err <<", status: "
			   << status << FEI_ENDL;
  }

  if (localProc==0) {
    FEI_COUT << " cpu-time in solve: " << solve_time << FEI_ENDL;
  }

  CHK_ERR( solnVec->scatterToOverlap() );

  delete hex_beam_ptr;

  double elapsed_cpu_time = fei::utils::cpu_time() - start_time;

  //The following IOS_... macros are defined in base/fei_iostream.hpp
  FEI_COUT.setf(IOS_FIXED, IOS_FLOATFIELD);
  if (localProc==0) {
    FEI_COUT << "Proc0 cpu times (seconds):" << FEI_ENDL
	 << "   FEI initialize:    " << fei_init_time << FEI_ENDL
         << "   FEI create-matrix: " << fei_creatematrix_time << FEI_ENDL
         << "   FEI load:          " << fei_load_time << FEI_ENDL
         << "      solve:          " << solve_time << FEI_ENDL
         << "Total program time:   " << elapsed_cpu_time << FEI_ENDL;

  }

  return(0);
}

void copy_teuchos_parameterlist(const Teuchos::ParameterList& paramlist,
                                fei::ParameterSet& paramset)
{
  Teuchos::ParameterList::ConstIterator
    iter = paramlist.begin(),
    iter_end = paramlist.end();

  for(; iter != iter_end; ++iter) {
    const Teuchos::ParameterEntry& param = paramlist.entry(iter);
    if (param.isType<std::string>()) {
      paramset.add(fei::Param(paramlist.name(iter).c_str(), Teuchos::getValue<std::string>(param).c_str()));
    }
    else if (param.isType<double>()) {
      paramset.add(fei::Param(paramlist.name(iter).c_str(), Teuchos::getValue<double>(param)));
    }
    else if (param.isType<int>()) {
      paramset.add(fei::Param(paramlist.name(iter).c_str(), Teuchos::getValue<int>(param)));
    }
    else if (param.isType<bool>()) {
      paramset.add(fei::Param(paramlist.name(iter).c_str(), Teuchos::getValue<bool>(param)));
    }
  }
}

