
#include "HexBeam.hpp"
#include "hexbeam_exception.hpp"
#include "hexbeam_functions.hpp"

#include <iostream>
#include <cmath>
#include <vector>
#include <sstream>

namespace hexbeam {

std::pair<int,int>
compute_num_procs_2D(int numProcsTotal, int W, int D)
{
  if (W == 0 || D == 0) {
    throw hexbeam::exception("compute_num_procs_2D ERROR, both W and D must be greater than 0");
  }

  std::vector<std::pair<int,int> > factors;

  factors.push_back(std::make_pair(1,numProcsTotal));

  for(int p=2; p<=numProcsTotal/2; ++p) {
    if (numProcsTotal%p == 0) {
      factors.push_back(std::make_pair(p,numProcsTotal/p));
    }
  }
  
  int offset = -1;
  int i = (int)factors.size()-1;
  while(i>=1) {
    if (W%factors[i].first == 0 && factors[i].second <= D) {
      offset = i;
    }
    --i;
  }

  if (offset < 0) {
    if (W%factors[0].first == 0 && factors[0].second <= D) {
      offset = 0;
    }
  }

  if (offset < 0) {
    std::ostringstream osstr;
    osstr << "compute_num_procs_2D ERROR: no feasible 2D\n"
       << "decomposition found for W="<<W<<", D="<<D
       <<" and numProcsTotal="<<numProcsTotal<<".\nNeed to find "
      << "numProcsW and numProcsD such that:\n"
      << " - numProcsW x numProcsD == numProcsTotal,\n"
      << " - W is an integer multiple of numProcsW,\n"
      << " - numProcsD <= D.\n";
    std::string str = osstr.str();
    throw hexbeam::exception(str);
  }

  return std::make_pair(factors[offset].first, factors[offset].second);
}

int print_beam_data(HexBeam& hexbeam, int numProcs, int localProc)
{
  FEI_COUT << localProc << ": num elems: " << hexbeam.numLocalElems() << FEI_ENDL;
  std::vector<int> nodeIDs(hexbeam.numNodesPerElem());
  std::vector<int> elemIDs;
  hexbeam.getLocalElemIDs(elemIDs);

  for(unsigned i=0; i<elemIDs.size(); ++i) {
    hexbeam.getElemConnectivity(elemIDs[i], &nodeIDs[0]);
    FEI_COUT << localProc << ": elem " << elemIDs[i] << ", nodes: ";
    for(int j=0; j<hexbeam.numNodesPerElem(); ++j) {
      FEI_COUT << nodeIDs[j] << " ";
    }
    FEI_COUT << FEI_ENDL;
  }

  return(0);
}


int init_elem_connectivities(fei::MatrixGraph* matrixGraph, HexBeam& hexbeam)
{
  std::vector<int> elemIDs;
  hexbeam.getLocalElemIDs(elemIDs);

  int numLocalElems = elemIDs.size();
  int nodesPerElem = hexbeam.numNodesPerElem();
  int fieldID = 0;
  int nodeIDType = 0;

  int patternID = 0;

  matrixGraph->definePattern(patternID, nodesPerElem, nodeIDType, fieldID);

  int blockID = 0;
  CHK_ERR(matrixGraph->initConnectivityBlock(blockID, numLocalElems, patternID));

  std::vector<int> nodeIDs(nodesPerElem);

  for(int i=0; i<numLocalElems; ++i) {
    CHK_ERR(hexbeam.getElemConnectivity(elemIDs[i], &nodeIDs[0]));

    CHK_ERR(matrixGraph->initConnectivity(blockID, elemIDs[i], &nodeIDs[0]));
  }

  return(0);
}

int init_shared_nodes(fei::MatrixGraph* matrixGraph, HexBeam& hexbeam)
{
  int numSharedNodes = hexbeam.getNumSharedNodes();
  if (numSharedNodes == 0) {
    return(0);
  }

  std::vector<int> sharedNodes;
  std::vector<int> numSharingProcsPerNode;
  std::vector<int> sharingProcs;

  if (numSharedNodes > 0) {
    CHK_ERR( hexbeam.getSharedNodes(sharedNodes, numSharingProcsPerNode,
				    sharingProcs) );
  }

  fei::SharedPtr<fei::VectorSpace> nodeSpace = matrixGraph->getRowSpace();

  int nodeIDType = 0;

  CHK_ERR( nodeSpace->initSharedIDs(sharedNodes.size(), nodeIDType,
				    &sharedNodes[0], &numSharingProcsPerNode[0],
				    &sharingProcs[0]) );

  return(0);
}

int init_constraints(fei::MatrixGraph* matrixGraph, HexBeam& hexbeam,
		     int localProc, int& firstLocalCRID)
{
  int numCRs = hexbeam.getNumCRs();
  if (numCRs < 1) {
    return(0);
  }

  int numNodesPerCR = hexbeam.getNumNodesPerCR();
  int* crnodes_1d = new int[numCRs*numNodesPerCR];
  int** crNodes = new int*[numCRs];
  int i, offset = 0;
  for(i=0; i<numCRs; ++i) {
    crNodes[i] = &(crnodes_1d[offset]);
    offset += numNodesPerCR;
  }

  CHK_ERR( hexbeam.getCRNodes(crNodes) );

  int crID = localProc*100000;
  firstLocalCRID = crID;

  int nodeIDType = 0;
  int crIDType = 1;

  std::vector<int> fieldIDs(numNodesPerCR, 0);
  std::vector<int> idTypes(numNodesPerCR, nodeIDType);

  for(i=0; i<numCRs; ++i) {
    CHK_ERR( matrixGraph->initLagrangeConstraint(crID+i, crIDType,
						 numNodesPerCR,
						 &idTypes[0], crNodes[i],
						 &fieldIDs[0]) );
  }

  delete [] crnodes_1d;
  delete [] crNodes;

  return(0);
}

int init_slave_constraints(fei::MatrixGraph* matrixGraph, HexBeam& hexbeam,
                           int dofPerNode)
{
  int numCRs = hexbeam.getNumCRs();
  if (numCRs < 1) {
    return(0);
  }

  int numNodesPerCR = hexbeam.getNumNodesPerCR();
  int* crnodes_1d = new int[numCRs*numNodesPerCR];
  int** crNodes = new int*[numCRs];
  int i, offset = 0;
  for(i=0; i<numCRs; ++i) {
    crNodes[i] = &(crnodes_1d[offset]);
    offset += numNodesPerCR;
  }

  CHK_ERR( hexbeam.getCRNodes(crNodes) );

  int nodeIDType = 0;

  int* fieldIDs = new int[numNodesPerCR];
  int* idTypes = new int[numNodesPerCR];
  for(i=0; i<numNodesPerCR; ++i) {
    fieldIDs[i] = 0;
    idTypes[i] = nodeIDType;
  }

  int fieldSize = dofPerNode;
  double* weights = new double[fieldSize*numNodesPerCR];

  for(i=0; i<fieldSize*numNodesPerCR; ++i) weights[i] = 0.0;
  weights[0] = -1.0;
  weights[fieldSize] = 1.0;
  double rhsValue = 0.0;
  int offsetOfSlave = 0;
  int offsetIntoSlaveField = 0;

  for(i=0; i<numCRs; ++i) {
    CHK_ERR( matrixGraph->initSlaveConstraint(numNodesPerCR,
					      idTypes,
					      crNodes[i],
					      fieldIDs,
					      offsetOfSlave,
					      offsetIntoSlaveField,
					      weights,
					      rhsValue) );
  }

  delete [] crnodes_1d;
  delete [] crNodes;
  delete [] fieldIDs;
  delete [] weights;

  return(0);
}

void elemStiffnessMatrix(int elemID,
                         int numNodes,
                         int dofPerNode,
                         std::vector<double>& stiffness)
{
  int len = numNodes * dofPerNode;
  stiffness.assign(len, 0.0);

  //Should set up some semi-realistic stiffness-matrix coefficients here...
  //For now just use arbitrary numbers and set it up so the matrix won't be
  //too ill-conditioned. (This is intended for an assembly test more than
  //a solver test.)

  //set the diagonal to 4.0
  for(int i=0; i<len; ++i) {
    int offset = i*len+i;
    stiffness[offset] = 4.0;
  }

  //Now set some off-diagonals
  for(int i=0; i<len; ++i) {
    int offset = i*len+i;
    if (i>1) {
      stiffness[offset-2] = -0.5;
    }

    if (i<len-2) {
      stiffness[offset+2] = -0.5;
    }

    if (i>3) {
      stiffness[offset-4] = -0.1;
    }
    if (i<len-4) {
      stiffness[offset+4] = -0.1;
    }
  }
}

void elemLoadVector(int elemID,
                    int numNodes,
                    int dofPerNode,
                    std::vector<double>& load)
{
  int len = numNodes*dofPerNode;
  load.resize(len);
  for(int i=0; i<len; ++i) {
    load[i] = 1.0;
  }
}

int load_elem_data(fei::MatrixGraph* matrixGraph,
		   fei::Matrix* mat,
		   fei::Vector* rhs,
		   HexBeam& hexbeam,
                   int dofPerNode)
{
  int blockID = 0;
  std::vector<int> elemIDs;
  hexbeam.getLocalElemIDs(elemIDs);
  int numLocalElems = elemIDs.size();
  int nodesPerElem = hexbeam.numNodesPerElem();

  int len = nodesPerElem*dofPerNode;
  std::vector<double> elemMat(len*len);
  double** elemMat2D = new double*[len];
  std::vector<double> elemVec(len);

  for(int j=0; j<len; ++j) {
    elemMat2D[j] = &(elemMat[j*len]);
  }

  elemStiffnessMatrix(elemIDs[0], hexbeam.numNodesPerElem(),
                      dofPerNode, elemMat);
  elemLoadVector(elemIDs[0], hexbeam.numNodesPerElem(), dofPerNode, elemVec);

  bool block_matrix = mat->usingBlockEntryStorage();
  std::vector<int> indices(len);

  if (block_matrix) {
    mat->getMatrixGraph()->setIndicesMode(fei::MatrixGraph::POINT_ENTRY_GRAPH);
  }

  for(int i=0; i<numLocalElems; ++i) {
    CHK_ERR( mat->getMatrixGraph()->getConnectivityIndices(blockID,
                                                           elemIDs[i],
                                                           len, &indices[0],
                                                           len) );
    CHK_ERR( mat->sumIn(len, &indices[0], len, &indices[0],
                        elemMat2D, FEI_DENSE_COL) );
    CHK_ERR( rhs->sumIn(len, &indices[0], &elemVec[0], 0) );
  }

  if (block_matrix) {
    mat->getMatrixGraph()->setIndicesMode(fei::MatrixGraph::BLOCK_ENTRY_GRAPH);
  }
  delete [] elemMat2D;

  return(0);
}

int load_constraints(fei::LinearSystem* linSys, HexBeam& hexbeam,
		     int dofPerNode, int firstLocalCRID)
{
  int numCRs = hexbeam.getNumCRs();
  if (numCRs < 1) {
    return(0);
  }

  int numNodesPerCR = hexbeam.getNumNodesPerCR();

  int fieldSize = dofPerNode;
  double* weights = new double[fieldSize*numNodesPerCR];

  int i;
  for(i=0; i<fieldSize*numNodesPerCR; ++i) weights[i] = 0.0;
  weights[0] = -1.0;
  weights[fieldSize] = 1.0;
  double rhsValue = 0.0;

  for(i=0; i<numCRs; ++i) {
    CHK_ERR( linSys->loadLagrangeConstraint(firstLocalCRID+i,
					    weights, rhsValue) );
  }

  delete [] weights;

  return(0);
}

int load_BC_data(fei::LinearSystem* linSys, HexBeam& hexbeam, int dofPerNode)
{
  int numBCNodes = hexbeam.getNumBCNodes();
  if (numBCNodes == 0) {
    return(0);
  }

  int* nodeIDs = new int[numBCNodes];
  if (nodeIDs == NULL) {
    return(0);
  }

  CHK_ERR( hexbeam.getBCNodes(numBCNodes, nodeIDs) );


  int fieldID = 0;
  int nodeIDType = 0;

  std::vector<double> bcvals(numBCNodes, 0.0);

  CHK_ERR( linSys->loadEssentialBCs(numBCNodes, nodeIDs,
				    nodeIDType, fieldID, 0,
				    &bcvals[0]) );

  delete [] nodeIDs;

  return(0);
}

}//namespace hexbeam

