#include "fei_iostream.hpp"

#include <sstream>

#include "utest_testcase1.hpp"

#include "hexbeam_exception.hpp"
#include "hexbeam_functions.hpp"

namespace beam_utest {

void check_hexbeam_decomp(int W, int D, HexBeam::Decomposition decomp,
                          int numProcs, int localProc,
                          int expect_numLocalElems,
                          int expect_numLocalNodes)
{
  HexBeam hexbeam(W, D, decomp, numProcs, localProc);

  if (hexbeam.numLocalElems() != expect_numLocalElems) {
    std::ostringstream osstr;
    osstr << "HexBeam("<<W<<","<<D<<","<<to_string(decomp)<<", "
        <<numProcs<<", "<<localProc
        <<").numLocalElems() == "<<hexbeam.numLocalElems()<<", should be "
      <<expect_numLocalElems<<".";
    std::string str = osstr.str();
    throw hexbeam::exception(str);
  }

  if (hexbeam.numLocalNodes() != expect_numLocalNodes) {
    std::ostringstream osstr;
    osstr << "HexBeam("<<W<<","<<D<<","<<to_string(decomp)<<", "
        <<numProcs<<", "<<localProc
        <<").numLocalNodes() == "<<hexbeam.numLocalNodes()<<", should be "
      <<expect_numLocalNodes<<".";
    std::string str = osstr.str();
    throw hexbeam::exception(str);
  }
}

void
testcase1::run(MPI_Comm comm)
{
  FEI_COUT << "testcase1: testing HexBeam decomposition...";

  int numProcs = 10;
  int localProc = 0;
  int W = 10;
  int D = 10;

  //for a "beam" of size 10x10x10 split in 1D on 10 processors:
  //each processor should have a 10x10x1 "slice" of elements
  //num-local-elements on proc 0 should be 100.
  //num-local-nodes on proc 0 should be 11x11x2 == 242

  check_hexbeam_decomp(W, D, HexBeam::OneD, numProcs, localProc,
                       100, 242);

  numProcs = 3;

  //for the same "beam" split on 3 processors:
  //num-local-elements on proc 0 should be 400
  //num-local-nodes on proc 0 should be 11x11x5 == 605

  check_hexbeam_decomp(W, D, HexBeam::OneD, numProcs, localProc,
                       400, 605);

  localProc = 2;

  //for the same "beam" split on 3 processors:
  //num-local-elements on proc 2 should be 300
  //num-local-nodes on proc 2 should be 11x11x4 == 484

  check_hexbeam_decomp(W, D, HexBeam::OneD, numProcs, localProc,
                       300, 484);

  numProcs = 100;

  //for the same "beam" split in 2D on 100 processors:
  //each processor should have a 10x1x1 mini-beam of elements.
  //num-local-elements on proc 2 should be 10x1 == 10
  //num-local-nodes on proc 2 should be 11x1x4 == 44

  check_hexbeam_decomp(W, D, HexBeam::TwoD, numProcs, localProc,
                       10, 44);

  numProcs = 50;
  localProc = 0;

  //for the same "beam" split in 2D on 50 processors:
  //each processor should have a 10x2x1 mini-slice of elements.
  //num-local-elements on proc 0 should be 20
  //num-local-nodes on proc 0 should be 11x3x2 == 66

  check_hexbeam_decomp(W, D, HexBeam::TwoD, numProcs, localProc,
                       20, 66);

  numProcs = 32;
  localProc = 0;
  W = 16;
  D = 2;

  //for a wide flat "beam of size 16x16x2 split in 2D on 32 processors:
  //each processor should have a 16x1x1 mini-slice of elements (i.e., the
  //decomposition should put 16 cuts across the W dimension, and 2 cuts
  //across the D dimension).
  //num-local-elements on proc 0 should be 16
  //num-local-nodes on proc 0 should be 17x2x2 == 68

  check_hexbeam_decomp(W, D, HexBeam::TwoD, numProcs, localProc,
                       16, 68);

  numProcs = 4;
  localProc = 0;
  W = 2;
  D = 2;

  //for a 2x2x2 cube of elements split in 2D on 4 processors:
  //each processor should have a 2x1x1 mini-beam of elements (i.e., the
  //decomposition should put 1 cut across the W dimension and 1 cut across
  //the D dimension.
  //num-local-elements on proc 0 should be 2.
  //num-local-nodes on proc 0 should be 3x2x2 = 12

  check_hexbeam_decomp(W, D, HexBeam::TwoD, numProcs, localProc,
                       2, 12);

  FEI_COUT << "ok"<<FEI_ENDL;
}

}//namespace beam_utest

