#include "fei_iostream.hpp"

#include <sstream>

#include "utest_testcase4.hpp"

#include "hexbeam_exception.hpp"
#include "hexbeam_functions.hpp"

namespace beam_utest {

void check_hexbeam_getSharedNodes(int W, int D, HexBeam::Decomposition decomp,
                          int numProcs, int localProc,
                          std::vector<int>& expect_nodeIDs,
                          std::vector<int>& expect_numProcsPerNode,
                          std::vector<int>& expect_sharingProcs)
{
  HexBeam hexbeam(W, D, decomp, numProcs, localProc, true);

  std::vector<int> nodeIDs;
  std::vector<int> numProcsPerNode;
  std::vector<int> sharingProcs;

  hexbeam.getSharedNodes(nodeIDs, numProcsPerNode, sharingProcs);

  if (nodeIDs != expect_nodeIDs || numProcsPerNode != expect_numProcsPerNode ||
      sharingProcs != expect_sharingProcs) {
    std::ostringstream osstr;
    osstr << "HexBeam("<<W<<","<<D<<","<<to_string(decomp)<<", "
        <<numProcs<<", "<<localProc
        <<").getSharedNodes(...) returned bad shared node info.\n"
      << "expected:\n";
    unsigned offset = 0;
    for(unsigned i=0; i<expect_nodeIDs.size(); ++i) {
      osstr << expect_nodeIDs[i] << ": procs: ";
      for(int j=0; j<expect_numProcsPerNode[i]; ++j) {
        osstr << expect_sharingProcs[offset++] << " ";
      }
      osstr << std::endl;
    }
    osstr << "\nreturned:\n";
    offset = 0;
    for(unsigned i=0; i<nodeIDs.size(); ++i) {
      osstr << nodeIDs[i] << ": procs: ";
      for(int j=0; j<numProcsPerNode[i]; ++j) {
        osstr << sharingProcs[offset++] << " ";
      }
      osstr << std::endl;
    }

    std::string str = osstr.str();
    throw hexbeam::exception(str);
  }
}

void
testcase4::run(MPI_Comm comm)
{
  FEI_COUT << "testcase4: testing HexBeam getSharedNodes...";

  int numProcs = 2;
  int localProc = 1;
  int W = 2;
  int D = 2;

  //for a "beam" of size 2x2x2 split in 1D on 2 processors:
  //each processor should have a 2x2x1 "slice" of elements
  //on proc 1, 9 nodes should be shared, and they should be
  //nodes 10 through 18

  std::vector<int> nodeIDs(9);
  for(unsigned i=0; i<9; ++i) nodeIDs[i] = i+10;
  std::vector<int> numProcsPerNode(9,1);
  std::vector<int> sharingProcs(9,0);

  check_hexbeam_getSharedNodes(W, D, HexBeam::OneD, numProcs, localProc,
                                nodeIDs, numProcsPerNode, sharingProcs);

  numProcs = 4;

  //for the same "beam" split in 2D on 4 processors:
  //each processor should have a 2x1x1 mini-beam of elements
  //on proc 1, 9 nodes should be shared: nodes 4,5,6 with proc 0,
  //nodes 13,14,15 with procs 0, 2, 3 and nodes 16,17,18 with proc 3.

  nodeIDs.resize(9);
  numProcsPerNode.resize(9);
  sharingProcs.resize(0);

  nodeIDs[0] = 4; numProcsPerNode[0] = 1; sharingProcs.push_back(0);
  nodeIDs[1] = 5; numProcsPerNode[1] = 1; sharingProcs.push_back(0);
  nodeIDs[2] = 6; numProcsPerNode[2] = 1; sharingProcs.push_back(0);

  nodeIDs[3] = 13; numProcsPerNode[3] = 3; sharingProcs.push_back(0);
    sharingProcs.push_back(2); sharingProcs.push_back(3);
  nodeIDs[4] = 14; numProcsPerNode[4] = 3; sharingProcs.push_back(0);
    sharingProcs.push_back(2); sharingProcs.push_back(3);
  nodeIDs[5] = 15; numProcsPerNode[5] = 3; sharingProcs.push_back(0);
    sharingProcs.push_back(2); sharingProcs.push_back(3);

  nodeIDs[6] = 16; numProcsPerNode[6] = 1; sharingProcs.push_back(3);
  nodeIDs[7] = 17; numProcsPerNode[7] = 1; sharingProcs.push_back(3);
  nodeIDs[8] = 18; numProcsPerNode[8] = 1; sharingProcs.push_back(3);

  check_hexbeam_getSharedNodes(W, D, HexBeam::TwoD, numProcs, localProc,
                                nodeIDs, numProcsPerNode, sharingProcs);

  numProcs = 6;
  localProc = 2;
  W = 2;
  D = 3;

  //for a 2x2x3 cube of elements split in 2D on 6 processors:
  //each processor should have a 2x1x1 mini-beam of elements.
  //on proc 2 there should be 12 shared nodes.
  //nodes 10, 11, 12 with proc 0
  //nodes 13, 14, 15 with procs 0, 1, 3
  //nodes 19, 20, 21 with proc 4
  //nodes 22, 23, 24 with procs 3, 4, 5

  nodeIDs.resize(12);
  numProcsPerNode.resize(12);
  sharingProcs.resize(0);

  nodeIDs[0] = 10; numProcsPerNode[0] = 1; sharingProcs.push_back(0);
  nodeIDs[1] = 11; numProcsPerNode[1] = 1; sharingProcs.push_back(0);
  nodeIDs[2] = 12; numProcsPerNode[2] = 1; sharingProcs.push_back(0);

  nodeIDs[3] = 13; numProcsPerNode[3] = 3; sharingProcs.push_back(0);
    sharingProcs.push_back(1); sharingProcs.push_back(3);
  nodeIDs[4] = 14; numProcsPerNode[4] = 3; sharingProcs.push_back(0);
    sharingProcs.push_back(1); sharingProcs.push_back(3);
  nodeIDs[5] = 15; numProcsPerNode[5] = 3; sharingProcs.push_back(0);
    sharingProcs.push_back(1); sharingProcs.push_back(3);

  nodeIDs[6] = 19; numProcsPerNode[6] = 1; sharingProcs.push_back(4);
  nodeIDs[7] = 20; numProcsPerNode[7] = 1; sharingProcs.push_back(4);
  nodeIDs[8] = 21; numProcsPerNode[8] = 1; sharingProcs.push_back(4);

  nodeIDs[9] = 22; numProcsPerNode[9] = 3; sharingProcs.push_back(3);
    sharingProcs.push_back(4); sharingProcs.push_back(5);
  nodeIDs[10] = 23; numProcsPerNode[10] = 3; sharingProcs.push_back(3);
    sharingProcs.push_back(4); sharingProcs.push_back(5);
  nodeIDs[11] = 24; numProcsPerNode[11] = 3; sharingProcs.push_back(3);
    sharingProcs.push_back(4); sharingProcs.push_back(5);

  check_hexbeam_getSharedNodes(W, D, HexBeam::TwoD, numProcs, localProc,
                               nodeIDs, numProcsPerNode, sharingProcs);


  numProcs = 4;
  localProc = 2;
  W = 4;
  D = 2;
 
  //for a 4x4x2 slab of elements split in 2D on 4 processors:
  //each processor should have a 4x2x1 mini-slab of elements.
  //on proc 2 there should be 5x3+5 == 20 shared nodes.
  //nodes 26-40 shared with proc 0
  //nodes 36-40 shared with procs 0, 1, 3
  //nodes 61-65 shared with proc 3

  nodeIDs.resize(20);
  numProcsPerNode.resize(20);
  sharingProcs.resize(0);
  for(unsigned i=0; i<10; ++i) {
    nodeIDs[i] = 26+i;
    numProcsPerNode[i] = 1;
    sharingProcs.push_back(0);
  }
  for(unsigned i=10; i<15; ++i) {
    nodeIDs[i] = 26+i;
    numProcsPerNode[i] = 3;
    sharingProcs.push_back(0);
    sharingProcs.push_back(1);
    sharingProcs.push_back(3);
  }
  for(unsigned i=15; i<20; ++i) {
    nodeIDs[i] = 61+i-15;
    numProcsPerNode[i] = 1;
    sharingProcs.push_back(3);
  }

  check_hexbeam_getSharedNodes(W, D, HexBeam::TwoD, numProcs, localProc,
                               nodeIDs, numProcsPerNode, sharingProcs);

  W = 2;
  D = 3;

  //for a 2x2x3 beam of elements split in 2D on 4 processors:
  //procs 0 and 1 should have a 2x1x2 mini-slab of elements,
  //procs 2 and 3 should have a 2x1x1 mini-slab of elements.
  //proc 2 should have 9 shared nodes.
  //nodes 19,20,21 shared with proc 0
  //nodes 22,23,24 shared with proc 0,1,3
  //nodes 31,32,33 shared with proc 3
 
  nodeIDs.resize(9);
  numProcsPerNode.resize(9);
  sharingProcs.resize(0);
  for(unsigned i=0; i<3; ++i) {
    nodeIDs[i] = i+19;
    numProcsPerNode[i] = 1;
    sharingProcs.push_back(0);
  }
  for(unsigned i=3; i<6; ++i) {
    nodeIDs[i] = i+19;
    numProcsPerNode[i] = 3;
    sharingProcs.push_back(0); sharingProcs.push_back(1);
    sharingProcs.push_back(3);
  }
  for(unsigned i=6; i<9; ++i) {
    nodeIDs[i] = i+25;
    numProcsPerNode[i] = 1;
    sharingProcs.push_back(3);
  }

  check_hexbeam_getSharedNodes(W, D, HexBeam::TwoD, numProcs, localProc,
                               nodeIDs, numProcsPerNode, sharingProcs);

  FEI_COUT << "ok"<<FEI_ENDL;
}

}//namespace beam_utest

